#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ddlpy` package."""

import pytest

import requests

from ddlpy import ddlpy


@pytest.fixture
def endpoints():
    """
    Get the endpoints from the api
    """
    return ddlpy.ENDPOINTS

@pytest.fixture
def collect_catalogue_resp(endpoints):
    endpoint = endpoints['collect_catalogue']
    resp = requests.post(endpoint['url'], json=endpoint['request'])
    return resp

def test_collect_catalogue(collect_catalogue_resp):
    assert collect_catalogue_resp.status_code == 200


@pytest.fixture
def collect_observations_resp(endpoints):
    endpoint = endpoints['collect_observations']
    resp = requests.post(endpoint['url'], json=endpoint['request'])
    return resp

def test_collect_observations(collect_observations_resp):
    assert collect_observations_resp.status_code == 200


@pytest.fixture
def collect_latest_observations_resp(endpoints):
    endpoint = endpoints['collect_latest_observations']
    resp = requests.post(endpoint['url'], json=endpoint['request'])
    return resp

def test_collect_latest_observations(collect_latest_observations_resp):
    assert collect_latest_observations_resp.status_code == 200


@pytest.fixture
def check_observations_available_resp(endpoints):
    endpoint = endpoints['check_observations_available']
    resp = requests.post(endpoint['url'], json=endpoint['request'])
    return resp

def test_check_observations_available(check_observations_available_resp):
    assert check_observations_available_resp.status_code == 200

@pytest.fixture
def collect_number_of_observations_resp(endpoints):
    endpoint = endpoints['collect_number_of_observations']
    resp = requests.post(endpoint['url'], json=endpoint['request'])
    return resp

def test_collect_number_of_observations(collect_number_of_observations_resp):
    assert collect_number_of_observations_resp.status_code == 200

@pytest.fixture
def request_bulk_observations_resp(endpoints):
    endpoint = endpoints['request_bulk_observations']
    resp = requests.post(endpoint['url'], json=endpoint['request'])
    return resp

def test_request_bulk_observations(request_bulk_observations_resp):
    assert request_bulk_observations_resp.status_code == 200



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 12 15:33:59 2024

@author: veenstra
"""

from ddlpy.waterinfo import waterinfo_read
import os

dir_tests = os.path.dirname(os.path.abspath(__file__))


def test_waterinfo_read():
    # example in May, during DST
    f = os.path.join(dir_tests, '20200608_069_20200507.csv')
    g = os.path.join(dir_tests, 'NVT_WATHTE_SCHE_20200507.csv')
    
    dxf_list = waterinfo_read(f)
    dxg_list = waterinfo_read(g)
    
    dxf0 = dxf_list[0]
    dxg0 = dxg_list[0]
    assert "time" in dxf0.variables
    assert "data" in dxf0.variables
    assert "time" in dxg0.variables
    assert "data" in dxg0.variables



# -*- coding: utf-8 -*-

import pandas as pd
import xarray as xr
from datetime import datetime, timedelta
import pytz


def waterinfo_read(f, encoding='latin'):
    """
    Load RWS csv data of https://waterinfo.rws.nl into xarray
    
    Can handle two types of files
    - direct download from https://waterinfo.rws.nl, first column is 'Datum'
    - request via https://waterinfo.rws.nl for link per email, first column is 'MONSTER_IDENTIFICATIE'
    If multiple variables are present, a list of xarrays is returned.
    
    An xarray that can be transformed into a dataframe:
    
    TIMEZONE
    The time is converted to 'UTC' from 
    * 'CET' (default) for direct downloads
    * 'UTC+1' (default) for request per email
    Passing a value in hours is also possible: tzone=.
    To ignore timezone conversion, use tzone=None.
    NB according to helpdeskwater: the timezone for downloads
    is always the Dutch legal time at the time of downloading
    (e.g. CET (UTC+2) when downloading between march and 
    october, and UTC+1) rest of the year) whereas the historic
    % data download in the website part (via "Download meer data" )
    are always UTC+1 (no Daylight Saving Time applied).
    
    Example:
    >> ds = waterinfo_read('x.csv',tzone='UTC+1')
    >> df = ds.to_dataframe()
    >> df.head()
    
    """
    
    dfall = pd.read_csv(f,delimiter=';', encoding=encoding)
    
    if 'WAARNEMINGDATUM' in dfall.keys():
        variablecolumn = 'GROOTHEID_ CODE'
        tzone = 1
        
        dfall = dfall.loc[dfall['KWALITEITSOORDEEL_CODE']=='Normale waarde']
        
    elif 'Datum' in dfall.keys():
        variablecolumn = 'Parameter'
        tzone = 'CET'
        
    variables = list(set(dfall[variablecolumn]))
    
    ds = []
    
    for variable in variables:
        
        df = dfall[dfall[variablecolumn]==variable]
        
        print(len(df),' rows for variable: ',variable)
    
        if 'WAARNEMINGDATUM' in df.keys():
            #MONSTER_IDENTIFICATIE;MEETPUNT_IDENTIFICATIE;TYPERING_OMSCHRIJVING;TYPERING_CODE;GROOTHEID_OMSCHRIJVING;GROOTHEID_ CODE;PARAMETER_OMSCHRIJVING;PARAMETER_ CODE;EENHEID_CODE;HOEDANIGHEID_OMSCHRIJVING;HOEDANIGHEID_CODE;COMPARTIMENT_OMSCHRIJVING;COMPARTIMENT_CODE;WAARDEBEWERKINGSMETHODE_OMSCHRIJVING;WAARDEBEWERKINGSMETHODE_CODE;WAARDEBEPALINGSMETHODE_OMSCHRIJVING;WAARDEBEPALINGSMETHODE_CODE;BEMONSTERINGSSOORT_OMSCHRIJVING;BEMONSTERINGSSOORT_CODE;WAARNEMINGDATUM;WAARNEMINGTIJD;LIMIETSYMBOOL;NUMERIEKEWAARDE;ALFANUMERIEKEWAARDE;KWALITEITSOORDEEL_CODE;STATUSWAARDE;OPDRACHTGEVENDE_INSTANTIE;MEETAPPARAAT_OMSCHRIJVING;MEETAPPARAAT_CODE;BEMONSTERINGSAPPARAAT_OMSCHRIJVING;BEMONSTERINGSAPPARAAT_CODE;PLAATSBEPALINGSAPPARAAT_OMSCHRIJVING;PLAATSBEPALINGSAPPARAAT_CODE;BEMONSTERINGSHOOGTE;REFERENTIEVLAK;EPSG;X;Y;ORGAAN_OMSCHRIJVING;ORGAAN_CODE;TAXON_NAME
            #;Scheveningen;;;Waterhoogte berekend;WATHTBRKD;;;cm;t.o.v. Normaal Amsterdams Peil;NAP;Oppervlaktewater;OW;;;Astronomische waterhoogte mbv harmonische analyse;other:F012;;;01-05-2020;00:00:00;;-44;;Normale waarde;Ongecontroleerd;RIKZMON_WAT;;;;;;;-999999999;NVT;25831;586550,994420996;5772806,43069697;;;

            t = [datetime.strptime(t,'%d-%m-%Y%H:%M:%S') for t in df['WAARNEMINGDATUM']+df['WAARNEMINGTIJD']]

            data = df['NUMERIEKEWAARDE']/1.

            key_units = 'EENHEID_CODE'

            keys2meta = ['MEETPUNT_IDENTIFICATIE',\
                        'GROOTHEID_OMSCHRIJVING','GROOTHEID_ CODE','EENHEID_CODE',\
                        'HOEDANIGHEID_OMSCHRIJVING','HOEDANIGHEID_CODE',\
                        'COMPARTIMENT_CODE','COMPARTIMENT_OMSCHRIJVING',\
                        'WAARDEBEPALINGSMETHODE_CODE','WAARDEBEPALINGSMETHODE_OMSCHRIJVING',\
                        'KWALITEITSOORDEEL_CODE','STATUSWAARDE','OPDRACHTGEVENDE_INSTANTIE',\
                        'EPSG','X','Y']        

        elif 'Datum' in df.keys():
            #Datum;Tijd;Parameter;Locatie;Meting;Verwachting;Astronomisch getijden;Eenheid;Bemonsteringshoogte;Referentievlak;
            #5-5-2020;21:10:00;Waterhoogte Oppervlaktewater t.o.v. Normaal Amsterdams Peil in cm;Den Helder;-1;;;cm;-999999999;NAP;

            # handle trailing ;
            df = df.loc[:, ~df.columns.str.contains('^Unnamed')]

            # parse time
            t = [datetime.strptime(t,'%d-%m-%Y%H:%M:%S') for t in df['Datum']+df['Tijd']]

            data = df['Meting']    
            key_units = 'Eenheid'

            keys2meta = ['Parameter','Locatie','Eenheid','Bemonsteringshoogte','Referentievlak'];

        else:
            print("Unknown file header.")
            raise

        if isinstance(tzone,type('a')):
            # does not apply DST t = [t1.replace(tzinfo=pytz.timezone(tzone)).astimezone(pytz.timezone('UTC')) for t1 in t]
            t = [t1 - pytz.timezone('CET').utcoffset(t1) for t1 in t]
        else:
            t = [t1 - timedelta(seconds=3600)*tzone for t1 in t]
        t = [t1.replace(tzinfo=None) for t1 in t] # make datetime64 again

        # The array values in a DataArray have a single (homogeneous) data type.
        # To work with heterogeneous or structured data types in xarray, use coordinates, 
        # or put separate DataArray objects in a single Dataset (see below).
        # ds2 = xr.DataArray(data, coords=[t], dims=['time'])

        d = xr.Dataset(
            {"data": (("time"), data)},     
            {'time': t}) # make datetime64 again
            
        d.attrs['file.name'] = f
        d.attrs['file.encoding'] = encoding
        d.attrs['file.original_columns'] = df.keys()
        d.attrs['file.original_timezone'] = tzone

        # unit conversoin to SI
        # LUT = {'in':['cm'],'out':['m'],'f':[0.01]}

        for key in keys2meta:
            value = set(df[key])
            if len(value) == 1:
                d['data'].attrs[key] = list(value)[0]
            else:
                d['data'].attrs[key] = value

        d['data'].attrs['units'] = d['data'].attrs[key_units]

        if tzone:
            d['time'].attrs['timezone'] = 'UTC'
            
        ds.append(d)
    
    if len(ds)==0:
        raise ValueError("no data available in file")
    
    return ds



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `utils` package."""

from ddlpy import utils
import datetime


def test_content():
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string
    start = datetime.datetime(2018, 1, 1)
    end = datetime.datetime(2018, 3, 1)
    result = utils.date_series(start, end)
    expected = [
        (datetime.datetime(2018, 1, 1, 0, 0), datetime.datetime(2018, 2, 1, 0, 0)),
        (datetime.datetime(2018, 2, 1, 0, 0), datetime.datetime(2018, 3, 1, 0, 0))
    ]
    assert result == expected

    start = datetime.datetime(2017, 11, 15)
    end = datetime.datetime(2018, 3, 5)
    result = utils.date_series(start, end)
    expected = [
        (datetime.datetime(2017, 11, 15, 0, 0), datetime.datetime(2017, 12, 15, 0, 0)),
        (datetime.datetime(2017, 12, 15, 0, 0), datetime.datetime(2018, 1, 15, 0, 0)),
        (datetime.datetime(2018, 1, 15, 0, 0), datetime.datetime(2018, 2, 15, 0, 0)),
        (datetime.datetime(2018, 2, 15, 0, 0), datetime.datetime(2018, 3, 5, 0, 0))
    ]
    assert result == expected



# -*- coding: utf-8 -*-

"""Main module."""
import json
import pathlib
import logging

import requests
import pandas as pd
import pytz
import tqdm
import dateutil
import numpy as np

from .utils import date_series

BASE_URL = "https://waterwebservices.rijkswaterstaat.nl/"
ENDPOINTS_PATH = pathlib.Path(__file__).with_name("endpoints.json")

with ENDPOINTS_PATH.open() as f:
    ENDPOINTS = json.load(f)


class NoDataError(ValueError):
    pass


class UnsuccessfulRequestError(ValueError):
    pass


# Web Feature Service
# Web Mapping Service
logger = logging.getLogger(__name__)


def _send_post_request(url, request, timeout=None):
    logger.debug("Requesting at {} with request: {}".format(url, json.dumps(request)))
    resp = requests.post(url, json=request, timeout=timeout)
    if not resp.ok:
        raise IOError("Request failed: {}".format(resp.text))
    
    result = resp.json()
    if not result['Succesvol']:
        logger.debug('Response result is unsuccessful: {}'.format(result))
        error_message = result.get('Foutmelding', 'No error returned')
        if error_message == "Geen gegevens gevonden!":
            # Foutmelding: "Geen gegevens gevonden!"
            # this is a valid response for periods where there is no data
            # this error is raised here, but catched in ddlpy.ddlpy.measurements() so the process can continue.
            raise NoDataError(error_message)
        else:
            # Foutmelding: "Het max aantal waarnemingen (157681) is overschreven, beperk uw request."
            # or any other possible error message are raised here
            raise UnsuccessfulRequestError(error_message)
    
    # continue if request was successful
    return result


def catalog(catalog_filter=None):
    endpoint = ENDPOINTS["collect_catalogue"]
    
    if catalog_filter is None:
        # use the default request from endpoints.json
        request = endpoint["request"]
    else:
        assert isinstance(catalog_filter, list)
        request = {"CatalogusFilter": {x:True for x in catalog_filter}}
    
    result = _send_post_request(endpoint["url"], request, timeout=None)
    
    return result


def locations(catalog_filter=None):
    """
    get station information from DDL (metadata from Catalogue). All metadata regarding stations.
    The response (result) retrieves more keys

    """

    result = catalog(catalog_filter=catalog_filter)
    
    df_locations = pd.DataFrame(result["LocatieLijst"])

    df_metadata = pd.json_normalize(result["AquoMetadataLijst"])

    df_metadata_location = pd.DataFrame(result["AquoMetadataLocatieLijst"])

    merged = (
        df_metadata_location.set_index("Locatie_MessageID")
        .join(df_locations.set_index("Locatie_MessageID"), how="inner")
        .reset_index()
    )
    merged = merged.set_index("AquoMetaData_MessageID").join(
        df_metadata.set_index("AquoMetadata_MessageID")
    )
    # set station id as index
    return merged.set_index("Code")


def _check_convert_dates(start_date, end_date, return_str=True):
    start_date = pd.Timestamp(start_date)
    end_date = pd.Timestamp(end_date)
    
    # check if timezones are the same
    assert start_date.tz == end_date.tz
    
    # set UTC timezone if tz is None
    if start_date.tz is None:
        start_date = pytz.UTC.localize(start_date)
    if end_date.tz is None:
        end_date = pytz.UTC.localize(end_date)
    
    if start_date > end_date:
        raise ValueError(f"start_date {start_date} is larger than end_date {end_date}")
    
    if return_str:
        start_date_str = start_date.isoformat(timespec='milliseconds')
        end_date_str = end_date.isoformat(timespec='milliseconds')
        return start_date_str, end_date_str
    else:
        return start_date, end_date


def _get_request_dicts(location):
    
    # generate aquometadata dict from location "*.Code" values
    key_list = [x.replace(".Code","") for x in location.index if x.endswith(".Code")]
    aquometadata_dict = {key:{"Code":location[f"{key}.Code"]} for key in key_list}
    
    # generate location dict from relevant values
    locatie_dict = {
        "X": location["X"],
        "Y": location["Y"],
        # assert code is used as index
        # TODO: use  a numpy  compatible json encoder in requests
        "Code": location.get("Code", location.name),
    }
    
    request_dicts = {"AquoMetadata": aquometadata_dict,
                     "Locatie": locatie_dict}
    return request_dicts


def measurements_available(location, start_date, end_date):
    """checks if there are measurements for location, for the period start_date, end_date
    gives None if check was unsuccesfull
    gives True/False if there are / are no measurement available
    """
    endpoint = ENDPOINTS['check_observations_available']

    start_date_str, end_date_str = _check_convert_dates(start_date, end_date, return_str=True)

    request_dicts = _get_request_dicts(location)

    request = {
        "AquoMetadataLijst": [request_dicts["AquoMetadata"]],
        "LocatieLijst": [request_dicts["Locatie"]],
        "Periode": {
            "Begindatumtijd": start_date_str,
            "Einddatumtijd": end_date_str
        }
    }

    result = _send_post_request(endpoint["url"], request, timeout=5)
    
    # continue if request was successful
    logger.debug('Got response: {}'.format(result))
    if result['WaarnemingenAanwezig'] == 'true' :
        return True
    else:
        return False  


def measurements_amount(location, start_date, end_date, period="Jaar"):
    """checks how much measurements are available for a location, for the period start_date, end_date
    returns a DataFrame with columns Groeperingsperiode and AantalMetingen
    possible for Jaar/Maand/Dag
    """
    # TODO: there are probably more Groeperingsperiodes accepted by ddl, but not supported by ddlpy yet
    accepted_period = ["Jaar","Maand","Dag"]
    if period not in accepted_period:
        raise ValueError(f"period should be one of {accepted_period}, not '{period}'")
    
    endpoint = ENDPOINTS['collect_number_of_observations']
    
    start_date_str, end_date_str = _check_convert_dates(start_date, end_date, return_str=True)

    request_dicts = _get_request_dicts(location)

    request = {
        "AquoMetadataLijst": [request_dicts["AquoMetadata"]],
        "LocatieLijst": [request_dicts["Locatie"]],
        "Groeperingsperiode": period,
        "Periode": {
            "Begindatumtijd": start_date_str,
            "Einddatumtijd": end_date_str
        }
    }

    result = _send_post_request(endpoint["url"], request, timeout=None)

    # continue if request was successful
    df_list = []
    for one in result['AantalWaarnemingenPerPeriodeLijst']:
        df = pd.json_normalize(one['AantalMetingenPerPeriodeLijst'])
        
        # combine columns to a period string
        df["Groeperingsperiode"] = df["Groeperingsperiode.Jaarnummer"].apply(lambda x: f"{x:04d}")
        if period in ["Maand", "Dag"]:
            df["Groeperingsperiode"] = (df["Groeperingsperiode"] + "-" + 
                                        df["Groeperingsperiode.Maandnummer"].apply(lambda x: f"{x:02d}"))
        if period in ["Dag"]:
            df["Groeperingsperiode"] = (df["Groeperingsperiode"] + "-" + 
                                        df["Groeperingsperiode.Dag"].apply(lambda x: f"{x:02d}"))
        
        # select columns from dataframe and append to list
        df = df.set_index("Groeperingsperiode")
        df = df[["AantalMetingen"]]
        df_list.append(df)
        
        # concatenate and sum duplicated index
        amount_all = pd.concat(df_list).sort_index()
        amount_all = amount_all.groupby(amount_all.index).sum()
        return amount_all


def _combine_waarnemingenlijst(result, location):
    assert "WaarnemingenLijst" in result
    
    # flatten the datastructure
    rows = []
    for waarneming in result["WaarnemingenLijst"]:
        for row in waarneming["MetingenLijst"]:
            # metadata is a list of 1 value, flatten it
            new_row = {}
            for key, value in row["WaarnemingMetadata"].items():
                new_key = "WaarnemingMetadata." + key
                new_val = value[0] if len(value) == 1 else value
                new_row[new_key] = new_val

            # add remaining data
            for key, val in row.items():
                if key == "WaarnemingMetadata":
                    continue
                new_row[key] = val

            # add metadata
            for key, val in list(waarneming["AquoMetadata"].items()):
                if isinstance(val, dict) and "Code" in val and "Omschrijving" in val:
                    # some values have a code/omschrijving pair, flatten them
                    new_key = key + ".Code"
                    new_val = val["Code"]
                    new_row[new_key] = new_val

                    new_key = key + ".Omschrijving"
                    new_val = val["Omschrijving"]
                    new_row[new_key] = new_val
                else:
                    new_row[key] = val
            rows.append(new_row)
    # normalize and return
    df = pd.json_normalize(rows)
    
    # add other info
    df["Code"] = location.get("Code", location.name)

    for name in [
        "Coordinatenstelsel",
        "Naam",
        "X",
        "Y",
        "Parameter_Wat_Omschrijving",
    ]:
        df[name] = location[name]

    # set NA value
    if "WaarnemingMetadata.KwaliteitswaardecodeLijst" in df.columns:
        bool_nan = df["WaarnemingMetadata.KwaliteitswaardecodeLijst"] == "99"
        if "Meetwaarde.Waarde_Numeriek" in df.columns:
            df.loc[bool_nan,"Meetwaarde.Waarde_Numeriek"] = np.nan
    
    try:
        df["time"] = pd.to_datetime(df["Tijdstip"])
        df = df.set_index("time")
    except KeyError:
        logger.exception(
            "Cannot add time variable time because variable Tijdstip is not found"
        )

    return df


def _measurements_slice(location, start_date, end_date):
    """get measurements for location, for the period start_date, end_date, use measurements instead"""
    endpoint = ENDPOINTS["collect_observations"]

    start_date_str, end_date_str = _check_convert_dates(start_date, end_date, return_str=True)

    request_dicts = _get_request_dicts(location)
    
    request = {
        "AquoPlusWaarnemingMetadata": {
            "AquoMetadata": request_dicts["AquoMetadata"]
            },
        "Locatie": request_dicts["Locatie"],
        "Periode": {"Begindatumtijd": start_date_str, 
                    "Einddatumtijd": end_date_str},
    }

    result = _send_post_request(endpoint["url"], request, timeout=None)

    df = _combine_waarnemingenlijst(result, location)
    return df


def _clean_dataframe(measurements):
    len_raw = len(measurements)
    # drop duplicate rows (preserves e.g. different Grootheden/Groeperingen at same timestep)
    measurements = measurements.drop_duplicates()
    
    # remove Tijdstip column, has to be done after drop_duplicates to avoid too much to be dropped
    measurements = measurements.drop("Tijdstip", axis=1, errors='ignore')
    
    # sort dataframe on time, ddl returns non-sorted data
    measurements = measurements.sort_index()
    ndropped = len_raw - len(measurements)
    logger.debug(f"{ndropped} duplicated values dropped")
    return measurements


def measurements(location, start_date, end_date, freq=dateutil.rrule.MONTHLY, clean_df=True):
    """
    Return measurements for the given location and time window (start_date, end_date)

    Parameters
    ----------
    location : pd.Series
        Single row of the `ddlpy.locations()` DataFrame.
    start_date : str, dt.datetime, pd.Timestamp
        Start of the retrieval period.
    end_date : str, dt.datetime, pd.Timestamp
        End of the retrieval period.
    freq : None, dateutil.rrule.MONTHLY, dateutil.rrule.YEARLY, etc., optional
        The frequency in which to divide the requested period (e.g. yearly or monthly).
        Can also be None, in which case the entire dataset will be retrieved at once.
        Please note that 10-minute measurements can often not be downloaded in yearly (or larger) chunks 
        since the DDL limits the responses to 157681 values and several stations have duplicated timesteps.
        In that case the query will fail with an error or timeout or just return an empty result (as if there was no data).
        In that case, the user should fallback to monthly chunks.
        This is significantly slower but it is also much more robust. The default is dateutil.rrule.MONTHLY.
    clean_df : bool, optional
        Whether to sort the dataframe and remove duplicate rows. The default is True.
    
    """
    
    if isinstance(location, pd.DataFrame):
        raise TypeError("The provided location is a pandas.DataFrame, but should be a pandas.Series, "
                        "supply only one location/row instead, for instance by doing 'location.iloc[0]'")
    
    start_date, end_date = _check_convert_dates(start_date, end_date, return_str=False)
    
    measurements = []

    # data_present = measurements_available(
    #         location, start_date=start_date, end_date=end_date)
    # if not data_present:
    #     # early return in case of no data
    #     logger.debug("no data found for this station and time extent")
    #     return
    
    if freq is None:
        date_series_iterator = tqdm.tqdm([(start_date, end_date)])
    else:
        date_series_iterator = tqdm.tqdm(
            date_series(start_date, end_date, freq=freq)
        )
    
    for (start_date_i, end_date_i) in date_series_iterator:
        try:
            measurement = _measurements_slice(
                location, start_date=start_date_i, end_date=end_date_i
            )
            measurements.append(measurement)
        except NoDataError:
            continue

    if len(measurements) == 0:
        # return empty dataframe in case of no data
        logger.debug("no data found for this station and time extent")
        return pd.DataFrame()
    
    measurements = pd.concat(measurements)

    if clean_df:
        measurements = _clean_dataframe(measurements)
    
    return measurements


def measurements_latest(location):
    """checks if there are measurements for location, for the period start_date, end_date
    gives None if check was unsuccesfull
    gives True/False if there are / are no measurement available
    """
    endpoint = ENDPOINTS['collect_latest_observations']

    request_dicts = _get_request_dicts(location)
    
    request = {"AquoPlusWaarnemingMetadataLijst":[{"AquoMetadata":request_dicts["AquoMetadata"]}],
               "LocatieLijst":[request_dicts["Locatie"]]
               }

    result = _send_post_request(endpoint["url"], request, timeout=5)
    
    # continue if request was successful
    df = _combine_waarnemingenlijst(result, location)
    return df



# -*- coding: utf-8 -*-

"""Top-level package for Data Distributie Laag. Service from Rijkswaterstaat for distributing water quantity data.."""

__author__ = """Fedor Baart"""
__email__ = 'fedor.baart@deltares.nl'
__version__ = '0.5.0'

from ddlpy.ddlpy import locations
from ddlpy.ddlpy import (measurements, 
                         measurements_latest, 
                         measurements_available, 
                         measurements_amount,
                         )
from ddlpy.utils import (simplify_dataframe,
                         dataframe_to_xarray)

__all__ = ['locations', 
           'measurements',
           'measurements_latest', 
           'measurements_available', 
           'measurements_amount',
           'simplify_dataframe',
           'dataframe_to_xarray'
           ]



# -*- coding: utf-8 -*-

"""
Console script for ddlpy.
    - ``ddlpy --help``
    - ``ddlpy locations --help``
    - ``ddlpy measurements --help``
"""
import sys
import logging
import click
import pandas as pd
import ddlpy


@click.group()
@click.option('-v', '--verbose', count=True)
def cli(verbose,  args=None):
    """Console script for ddlpy."""
    level = logging.INFO
    if verbose >= 1:
        level = logging.DEBUG
    logging.basicConfig(level=level)
    return 0


# Define a command
# Each command has options which are read from the console.
@cli.command()
@click.option(
    '--output', 
    help='the locations json filename that will be created',
    default='locations.json'
    )
@click.option(
    '--station',
    help='Station codes, e.g. HOEKVHLD',
    multiple=True
)
@click.option(
    '--grootheid-code',
    help='Grootheid code, e.g. WATHTE',
    multiple=True
)
@click.option(
    '--groepering-code',
    help='Groepering code, e.g. NVT',
    multiple=True
)
@click.option(
    '--hoedanigheid-code',
    help='Hoedanigheid code, e.g. NAP',
    multiple=True
)
@click.option(
    '--eenheid-code',
    help='Eenheid code, e.g. cm',
    multiple=True
)
@click.option(
    '--parameter-code',
    help='Parameter code',
    multiple=True
)
@click.option(
    '--compartment-code',
    help='Compartment code, e.g. OW',
    multiple=True
)
def locations(output,
              station,
              grootheid_code,
              groepering_code,
              hoedanigheid_code,
              eenheid_code,
              parameter_code,
              compartment_code):
    """
    Subset locations dataframe based on input codes and write locations.json.

    """
    locations_df = ddlpy.locations()

    stations = station
    quantities = {'Grootheid.Code': list(grootheid_code),
                  'Groepering.Code': list(groepering_code),
                  'Hoedanigheid.Code': list(hoedanigheid_code),
                  'Eenheid.Code': list(eenheid_code),
                  'Parameter.Code': list(parameter_code),
                  'Compartiment.Code': list(compartment_code)
                  }

    selected = locations_df.copy()

    if (stations):
        selected = selected[selected.index.isin(stations)]

    for q in quantities.keys():
        if (len(quantities[q]) != 0):
            selected = selected[selected[q].isin(quantities[q])]

    selected.reset_index(inplace= True)

    output= output.split('.')[0] # make sure that extension is always json
    selected.to_json(output+'.json', orient='records')

# Another command to get the measurements from locations
@cli.command()
@click.argument(
    'start-date',
)
@click.argument(
    'end-date',
)
@click.option(
    '--locations',
    default='locations.json',
    help='file in json or parquet format containing locations and codes'
)
def measurements(locations, start_date, end_date):
    """
    Obtain measurements from file with locations and codes. 
    The arguments start_date and end_date should be formatted 
    like "YYYY-MM-DD" or something else that `pandas.Timestamp` understands.
    """
    try:
        locations_df = pd.read_json(locations, orient='records')
    except:
        raise ValueError('locations.json file not found. First run "ddlpy locations"')
        
    for obs in range(locations_df.shape[0]): #goes through rows in table
        selected = locations_df.loc[obs]

        measurements = ddlpy.measurements(
            selected, start_date=start_date, end_date=end_date)

        if (len(measurements) > 0):
            print('Measurements of %s were obtained' % selected['Code'])
            station = selected['Code']
            cc = selected['Compartiment.Code']
            ec = selected['Eenheid.Code']
            gc = selected['Grootheid.Code']
            grc = selected['Groepering.Code']
            hc = selected['Hoedanigheid.Code']
            pc = selected['Parameter.Code']

            measurements.to_csv('%s_%s_%s_%s_%s_%s_%s.csv' %
                                (station, cc, ec, gc, grc, hc, pc))
        else:
            print('No Data of station %s were retrieved from Water Info' %
                  selected['Code'])


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `ddlpy` package."""
import datetime as dt
import pandas as pd
import pytest
import ddlpy
import dateutil
import numpy as np


@pytest.fixture
def locations():
    """return all locations"""
    locations = ddlpy.locations()
    return locations


@pytest.fixture
def location(locations):
    """return sample location"""
    bool_grootheid = locations['Grootheid.Code'] == 'WATHTE'
    bool_groepering = locations['Groepering.Code'] == 'NVT'
    location = locations[bool_grootheid & bool_groepering].loc['DENHDR']
    return location


@pytest.fixture
def measurements(location):
    """measurements for a location """
    start_date = dt.datetime(1953, 1, 1)
    end_date = dt.datetime(1953, 4, 1)
    measurements = ddlpy.measurements(location, start_date=start_date, end_date=end_date)
    return measurements


def test_locations(locations):
    # the number of columns depend on the catalog filter in endpoints.json
    assert locations.shape[1] == 18
    # the number of rows is the number of stations, so will change over time
    assert locations.shape[0] > 1


def test_locations_extended():
    catalog_filter = ['Compartimenten','Eenheden','Grootheden',
                      'Hoedanigheden','Groeperingen','MeetApparaten',
                      'Typeringen','WaardeBepalingsmethoden','Parameters']
    locations_extended = ddlpy.locations(catalog_filter=catalog_filter)
    # the number of columns depend on the provided catalog_filter
    assert locations_extended.shape[1] == 24
    # the number of rows is the number of stations, so will change over time
    assert locations_extended.shape[0] > 1


def test_measurements(measurements):
    assert measurements.shape[0] > 1


def test_measurements_freq_yearly(location, measurements):
    start_date = dt.datetime(1953, 1, 1)
    end_date = dt.datetime(1953, 4, 1)
    measurements_yearly = ddlpy.measurements(location, start_date=start_date, end_date=end_date, freq=dateutil.rrule.YEARLY)
    assert measurements.shape == measurements_yearly.shape


def test_measurements_freq_none(location, measurements):
    start_date = dt.datetime(1953, 1, 1)
    end_date = dt.datetime(1953, 4, 1)
    measurements_yearly = ddlpy.measurements(location, start_date=start_date, end_date=end_date, freq=None)
    assert measurements.shape == measurements_yearly.shape


def test_measurements_available(location):
    start_date = dt.datetime(1953, 1, 1)
    end_date = dt.datetime(1953, 4, 1)
    data_present = ddlpy.measurements_available(location, start_date=start_date, end_date=end_date)
    assert isinstance(data_present, bool)


def test_measurements_amount(location):
    start_date = dt.datetime(1953, 1, 1)
    end_date = dt.datetime(1953, 4, 5)
    data_amount_dag = ddlpy.measurements_amount(location, start_date=start_date, end_date=end_date, period="Dag")
    assert data_amount_dag.shape[0] > 50
    assert data_amount_dag.index.str.len()[0] == 10
    data_amount_maand = ddlpy.measurements_amount(location, start_date=start_date, end_date=end_date, period="Maand")
    assert data_amount_maand.shape[0] == 4
    assert data_amount_maand.index.str.len()[0] == 7
    data_amount_jaar = ddlpy.measurements_amount(location, start_date=start_date, end_date=end_date, period="Jaar")
    assert data_amount_jaar.shape[0] == 1
    assert data_amount_jaar.index.str.len()[0] == 4


def test_measurements_latest(location):
    """measurements for a location """
    latest = ddlpy.measurements_latest(location)
    assert latest.shape[0] > 1


def test_measurements_empty(location):
    """measurements for a location """
    start_date = dt.datetime(2153, 1, 1)
    end_date = dt.datetime(2153, 1, 2)
    measurements = ddlpy.measurements(location, start_date=start_date, end_date=end_date)
    assert measurements.empty


def test_measurements_typerror(locations):
    start_date = dt.datetime(1953, 1, 1)
    end_date = dt.datetime(1953, 4, 1)
    with pytest.raises(TypeError):
        _ = ddlpy.measurements(locations, start_date=start_date, end_date=end_date)


def test_measurements_noindex(location):
    # pandas dataframe with Code as column instead of index
    locations_noindex = pd.DataFrame(location).T
    locations_noindex.index.name = "Code"
    locations_noindex = locations_noindex.reset_index(drop=False)
    
    # normal subsetting and retrieving
    location_sel = locations_noindex.iloc[0]
    start_date = dt.datetime(1953, 1, 1)
    end_date = dt.datetime(1953, 4, 1)
    measurements = ddlpy.measurements(location_sel, start_date=start_date, end_date=end_date)
    assert measurements.shape[0] > 1


def test_measurements_long(location):
    """measurements for a location """
    start_date = dt.datetime(1951, 11, 1)
    end_date = dt.datetime(1953, 4, 1)
    measurements = ddlpy.measurements(location, start_date=start_date, end_date=end_date)
    assert measurements.shape[0] > 1


def test_measurements_sorted(measurements):
    """https://github.com/deltares/ddlpy/issues/27"""
    
    # sort dataframe on values so it will not be sorted on time
    meas_wrongorder = measurements.sort_values("Meetwaarde.Waarde_Numeriek")
    assert meas_wrongorder.index.is_monotonic_increasing == False
    meas_clean = ddlpy.ddlpy._clean_dataframe(meas_wrongorder)
    assert meas_clean.index.is_monotonic_increasing == True
    # assert meas_clean.index.duplicated().sum() == 0
    
    # check wheter indexes are DatetimeIndex
    assert isinstance(meas_wrongorder.index, pd.DatetimeIndex)
    assert isinstance(meas_clean.index, pd.DatetimeIndex)


def test_measurements_duplicated(measurements):
    """
    WALSODN 2010 contains all values three times, ddlpy drops duplicates
    https://github.com/deltares/ddlpy/issues/24
    if the data is cleaned in ddl, this test will fail and can be removed or adjusted
    
    length assertion of meas_clean is important, to prevent issue 
    https://github.com/deltares/ddlpy/issues/53
    """
    
    # deliberately duplicate values in a measurements dataframe
    meas_duplicated = pd.concat([measurements, measurements, measurements], axis=0)
    meas_clean = ddlpy.ddlpy._clean_dataframe(meas_duplicated)
    assert len(meas_duplicated) == 3024
    assert len(meas_clean) == 392
    
    # check wheter indexes are DatetimeIndex
    assert isinstance(meas_duplicated.index, pd.DatetimeIndex)
    assert isinstance(meas_clean.index, pd.DatetimeIndex)


def test_measurements_timezone_behaviour(location):
    start_date = "2015-01-01 00:00:00 +01:00"
    end_date = "2015-01-03 00:00:00 +01:00"
    measurements = ddlpy.measurements(location, start_date=start_date, end_date=end_date)
    assert str(measurements.index[0].tz) == 'UTC+01:00'
    assert measurements.index[0] == pd.Timestamp(start_date)
    assert measurements.index[-1] == pd.Timestamp(end_date)
    
    data_amount_dag = ddlpy.measurements_amount(location, start_date=start_date, end_date=end_date, period="Dag")
    # when retrieving with tzone +01:00 we expect 1 value on 2015-01-03
    assert np.allclose(data_amount_dag["AantalMetingen"].values, [144,144,1])
    
    
    start_date = "2015-01-01"
    end_date = "2015-01-03"
    measurements = ddlpy.measurements(location, start_date=start_date, end_date=end_date)
    assert str(measurements.index[0].tz) == 'UTC+01:00'
    assert measurements.index[0] == pd.Timestamp(start_date).tz_localize("UTC").tz_convert('UTC+01:00')
    assert measurements.index[-1] == pd.Timestamp(end_date).tz_localize("UTC").tz_convert('UTC+01:00')
    
    data_amount_dag = ddlpy.measurements_amount(location, start_date=start_date, end_date=end_date, period="Dag")
    # when retrieving with tzone +00:00 we expect 7 values on 2015-01-03
    assert np.allclose(data_amount_dag["AantalMetingen"].values, [138,144,7])


def test_nodataerror(location):
    """
    Test whether a request that returns no data is indeed properly catched
    This is important since it is derived from the returned error message "Geen gegevens gevonden!"
    In case this error message changes in the future, 
    this test will fail and the ddlpy code needs to be updated accordingly
    """
    start_date = dt.datetime(2180, 1, 1)
    end_date = dt.datetime(2180, 4, 1)
    with pytest.raises(ddlpy.ddlpy.NoDataError):
        # ddlpy.measurements() catches NoDataError, so we have to test it with _measurements_slice
        _ = ddlpy.ddlpy._measurements_slice(location, start_date=start_date, end_date=end_date)
    with pytest.raises(ddlpy.ddlpy.NoDataError):
        _ = ddlpy.ddlpy.measurements_amount(location, start_date=start_date, end_date=end_date)


# TODO: this testcase is very slow and does not add much value, uncomment it when the ddl is faster
# def test_unsuccessfulrequesterror(location):
#     """
#     deliberately send a request that is too large to get the error message
#     Foutmelding: 'Het max aantal waarnemingen (157681) is overschreven, beperk uw request.'
#     which is raised as a UnsuccessfulRequestError
#     """
#     start_date = dt.datetime(2015, 1, 1)
#     end_date = dt.datetime(2020, 1, 1)
#     with pytest.raises(ddlpy.ddlpy.UnsuccessfulRequestError):
#         #this is the same as ddlpy.measurements(location, start_date=start_date, end_date=end_date, freq=None)
#         _ = ddlpy.ddlpy._measurements_slice(location, start_date=start_date, end_date=end_date)


datetype_list = ["string", "pd.Timestamp", "dt.datetime", "mixed"]
@pytest.mark.parametrize("datetype", datetype_list)
def test_check_convert_dates(datetype):
    if datetype == "string":
        start_date = "1953-01-01"
        end_date = "1953-04-01"
    elif datetype == "pd.Timestamp":
        start_date = pd.Timestamp("1953-01-01")
        end_date = pd.Timestamp("1953-04-01")
    elif datetype == "dt.datetime":
        start_date = dt.datetime(1953,1,1)
        end_date = dt.datetime(1953,4,1)
    elif datetype == "mixed":
        start_date = "1953-01-01"
        end_date = dt.datetime(1953,4,1)

    # assert output
    start_date_out, end_date_out = ddlpy.ddlpy._check_convert_dates(start_date, end_date)
    assert start_date_out=='1953-01-01T00:00:00.000+00:00'
    assert end_date_out=='1953-04-01T00:00:00.000+00:00'


def test_check_convert_wrongorder():
    start_date = "1953-01-01"
    end_date = "1953-04-01"
    
    # assert output
    with pytest.raises(ValueError):
        start_date_out, end_date_out = ddlpy.ddlpy._check_convert_dates(end_date, start_date)


def test_simplify_dataframe(measurements):
    assert len(measurements.columns) == 53
    meas_simple = ddlpy.simplify_dataframe(measurements)
    assert hasattr(meas_simple, "attrs")
    assert len(meas_simple.attrs) == 50
    assert len(meas_simple.columns) == 3


def test_dataframe_to_xarray(measurements):
    drop_if_constant = ["WaarnemingMetadata.OpdrachtgevendeInstantieLijst",
                        "WaarnemingMetadata.BemonsteringshoogteLijst",
                        "WaarnemingMetadata.ReferentievlakLijst",
                        "AquoMetadata_MessageID", 
                        "BemonsteringsSoort.Code", 
                        "Compartiment.Code", "Eenheid.Code", "Grootheid.Code", "Hoedanigheid.Code",
                        ]
    ds_clean = ddlpy.dataframe_to_xarray(measurements, drop_if_constant)
    
    # check if constant value that was not in drop_if_constant list is indeed not dropped
    assert "MeetApparaat.Code" in ds_clean.data_vars
    assert len(ds_clean["MeetApparaat.Code"]) > 0
    
    for varname in drop_if_constant:
        if varname == "WaarnemingMetadata.OpdrachtgevendeInstantieLijst":
            continue
        assert varname not in ds_clean.data_vars
        assert varname in ds_clean.attrs.keys()
    assert "WaarnemingMetadata.OpdrachtgevendeInstantieLijst" in ds_clean.data_vars
    assert "WaarnemingMetadata.OpdrachtgevendeInstantieLijst" not in ds_clean.attrs.keys()
    
    data_vars_list = ['WaarnemingMetadata.StatuswaardeLijst',
     'WaarnemingMetadata.KwaliteitswaardecodeLijst',
     'MeetApparaat.Code',
     'WaardeBepalingsmethode.Code',
     'Meetwaarde.Waarde_Numeriek']
    for varname in data_vars_list:
        assert varname in ds_clean.data_vars
    
    assert "X" in ds_clean.attrs.keys()



import dateutil.rrule
import itertools
import pandas as pd


def date_series(start, end, freq=dateutil.rrule.MONTHLY):
    """return a list of start and end date over the timespan start[->end following the frequency rule"""
    def pairwise(it):
        """return all sequential pairs"""
        # loop over the iterator twice.
        # tee it so we don't consume it twice
        it0, it1 = itertools.tee(it)
        i0 = itertools.islice(it0, None)
        i1 = itertools.islice(it1, 1, None)
        # merge to a list of pairs
        return zip(i0, i1)

    # go over the rrule, also include the end, return consequitive pairs
    result = list(
        pairwise(
            list(
                dateutil.rrule.rrule(dtstart=start, until=end, freq=freq)
            ) + [end]
        )
    )
    # remove last one if empty (first of month until first of month)
    if len(result) > 1 and result[-1][0] == result[-1][1]:
        # remove it
        del result[-1]
    return result


def simplify_dataframe(df: pd.DataFrame):
    """
    Drop columns with constant values from the dataframe and collect them 
    in a dictionary which is added as attrs of the dataframe.
    """
    
    bool_constant = (df == df.iloc[0]).all()
    
    # constant columns are flattened and converted to dict of attrs
    df_attrs = df.loc[:, bool_constant].iloc[0].to_dict()
    
    # varying columns are kept in output dataframe
    df_simple = df.loc[:, ~bool_constant].copy()
    
    # attach as attrs to dataframe
    df_simple.attrs = df_attrs
    
    return df_simple


def code_description_attrs_from_dataframe(df: pd.DataFrame):
    # create var_attrs_dict
    colname_code_list = df.columns[df.columns.str.contains(".Code")]
    var_attrs_dict = {}
    for colname_code in colname_code_list:
        colname_oms = colname_code.replace(".Code",".Omschrijving")
        meas_twocol = df[[colname_code,colname_oms]].drop_duplicates()
        attr_dict = meas_twocol.set_index(colname_code)[colname_oms].to_dict()
        var_attrs_dict[colname_code] = attr_dict
    return var_attrs_dict


def dataframe_to_xarray(df: pd.DataFrame, drop_if_constant=[]):
    """
    Converts the measurement dataframe to a xarray dataset,
    including several cleanups to minimize the size of the netcdf dataset on disk:
    
    - The column 'Parameter_Wat_Omschrijving' is dropped (combination of information in other columns)
    - The column 'Meetwaarde.Waarde_Alfanumeriek' is dropped if 'Meetwaarde.Waarde_Numeriek' is present (contains duplicate values in that case)
    - All Omschrijving columns are dropped and added as attributes to the Code variables
    - All NVT-only Code columns are dropped and added as ds attributes
    - All location columns are dropped and added as ds attributes
    - All drop_if_constant columns are dropped and added as ds attributes (if the values are indeed constant)
    
    The timestamps are converted to UTC since xarray does not support non-UTC timestamps.
    These can be converted to different timezones after loading the netcdf and converting 
    to a pandas dataframe with df.index.tz_convert()
    """

    # create list of columns with duplicate info (often not constant), will be dropped
    cols_bulky = ["Parameter_Wat_Omschrijving"]
    if "Meetwaarde.Waarde_Alfanumeriek" in df.columns and 'Meetwaarde.Waarde_Numeriek' in df.columns:
        # drop alfanumeriek if duplicate of numeriek # TODO: should not be returned by ddl
        cols_bulky.append("Meetwaarde.Waarde_Alfanumeriek")
    
    # create list of all omschrijving columns, will be dropped (added as ds[varn].attrs via code_description_attrs_from_dataframe())
    cols_omschrijving = df.columns[df.columns.str.contains(".Omschrijving")].tolist()
    
    # create list of all-NVT *.Code columns, will be dropped (codes added as ds.attrs)
    bool_onlynvt_code = (df=='NVT').all(axis=0)
    cols_onlynvt_code = df.columns[bool_onlynvt_code].tolist()
    cols_onlynvt_code = [x for x in cols_onlynvt_code if x.endswith(".Code")]
    
    # create list of location columns, will be dropped (added as ds.attrs)
    cols_location = ['Code', 'Naam', 'Coordinatenstelsel', 'X', 'Y']
    
    # add drop_if_constant colums to list if values are indeed constant, will be dropped (added as ds.attrs)
    cols_constant = []
    for colname in drop_if_constant:
        assert colname in df.columns
        if len(df[colname].drop_duplicates()) == 1:
            cols_constant.append(colname)
    
    # create ds attrs for all nvt/location/constant columns
    ds_attrs = {}
    attrs_columns = cols_onlynvt_code + cols_constant + cols_location
    for colname in attrs_columns:
        ds_attrs[colname] = df[colname].iloc[0]
    
    # drop columns 
    drop_columns = (cols_bulky + cols_location + cols_constant +
                    cols_onlynvt_code + cols_omschrijving)
    df_simple = df.drop(drop_columns, axis=1, errors='ignore')
    
    # convert to UTC to please xarray
    # TODO: adding tzone to time.encoding['units'] raises "ValueError: invalid time units: 1970-01-01 00:00:00 +01:00"
    df_simple.index = df_simple.index.tz_convert(None)
    
    # convert to xarray dataset and add ds_attrs
    ds = df_simple.to_xarray()
    ds = ds.assign_attrs(ds_attrs)

    # assign attrs with code+omschrijving to each *.Code variable
    var_attrs_dict = code_description_attrs_from_dataframe(df)
    for varn in ds.data_vars:
        if varn in var_attrs_dict.keys():
            var_attrs = var_attrs_dict[varn]
            ds[varn] = ds[varn].assign_attrs(var_attrs)

    return ds



# -*- coding: utf-8 -*-
"""
Created on Wed Mar 13 07:44:47 2024

@author: veenstra
"""

import os
from click.testing import CliRunner
from ddlpy import cli

def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.cli)
    assert result.exit_code == 0
    assert 'Show this message and exit.' in result.output
    help_result = runner.invoke(cli.cli, ['--help'])
    assert help_result.exit_code == 0
    assert 'Show this message and exit.' in help_result.output

    locations_command = 'locations --grootheid-code WATHTE --station HOEKVHLD'
    locations_result = runner.invoke(cli.cli, locations_command.split())
    assert locations_result.exit_code == 0
    file_locs = "locations.json"
    assert os.path.exists(file_locs)
    
    measurements_command = 'measurements 2023-01-01 2023-01-03'
    measurements_result = runner.invoke(cli.cli, measurements_command.split())
    assert measurements_result.exit_code == 0
    file_meas = "HOEKVHLD_OW_cm_WATHTE_NVT_NAP_NVT.csv"
    assert os.path.exists(file_meas)
    
    # cleanup
    os.remove(file_locs)
    os.remove(file_meas)
